from dataiku.core.dataset import Dataset
from dataiku.core.managed_folder import Folder
from dataiku.core.saved_model import Model
from dataikuapi.dss.dataset import DSSDataset
from dataikuapi.dss.managedfolder import DSSManagedFolder
from dataikuapi.dss.project import DSSProject
from dataikuapi.dss.savedmodel import DSSSavedModel

from dku_utils.deployed_models.deployed_model_commons import get_deployed_model_id
from dku_utils.folders.folder_commons import get_managed_folder_id


def compute_object_to_share_settings(
    object_name: str, object_type: str, rules: list, quick_share: bool = False
):
    """
    Structures a dictionary with the information for element sharing
    :param: object_name: str: Name of element to share
    :param: object_type: str: Type of the element to share
    :param: rules: list: List of the sharing rules
    :param: quick_share: bool: Is the quick sharing activated for this element

    :returns: object_sharing_dict dict: Setting of the element sharing for this element
    """
    object_sharing_dict = {
        "type": object_type,
        "localName": object_name,
        "quickSharingEnabled": quick_share,
        "rules": rules,
    }
    return object_sharing_dict


def get_object_to_share_type(obj):
    """
    Get the object type and id
    :param: obj: Element to extract type and id
    :returns: object_type str: Type of the object
    :returns: object_id str: ID of the object
    """
    if isinstance(obj, DSSDataset):
        return "DATASET", obj.id
    elif isinstance(obj, Dataset):
        return "DATASET", obj.short_name
    elif isinstance(obj, DSSManagedFolder):
        return "MANAGED_FOLDER", obj.id
    elif isinstance(obj, Folder):
        return "MANAGED_FOLDER", obj.get_id()
    elif isinstance(obj, Model):
        return "SAVED_MODEL", obj.get_id()
    elif isinstance(obj, DSSSavedModel):
        return "SAVED_MODEL", obj.id
    else:
        raise TypeError(f"{type(obj)} is not a recognized type for object sharing")


def share_object_from_id(
    object_id: str,
    object_type: str,
    current_project: DSSProject,
    target_project_key: str,
    quick_hare=False,
):
    """
    Share an object base on their id and type to a targeted
    :param: object_id str: ID of the element to share
    :param: object_type str: Type of the element to share
    :param: current_project DSSProject: Current project where the element is originated
    :param: target_project_key str: Project key of the target dataset
    :param: quick_hare bool : Is the quick sharing activated for this element
    """
    project_settings = current_project.get_settings()
    previous_shared_objects = project_settings.get_raw()["exposedObjects"]["objects"]

    # we check if this specific object is already shared to avoid erasing previous sharing
    matching_shared_objects = list(
        filter(
            lambda dct: (dct["localName"] == object_id)
            and (dct["type"] == object_type),
            previous_shared_objects,
        )
    )
    unmatching_shared_objects = list(
        filter(
            lambda dct: (dct["localName"] != object_id) or (dct["type"] != object_type),
            project_settings.get_raw()["exposedObjects"]["objects"],
        )
    )

    if len(matching_shared_objects) > 0:
        # there's previous sharing
        base_rules = matching_shared_objects[0]["rules"]
        already_shared_objects = list(
            filter(lambda dct: dct["targetProject"] == target_project_key, base_rules)
        )
        if len(already_shared_objects) > 0:
            # this object is already shared to the target project
            return None
    else:
        base_rules = []

    new_rule = {"targetProject": target_project_key, "appearOnFlow": True}
    base_rules.append(new_rule)

    object_sharing = compute_object_to_share_settings(
        object_name=object_id,
        object_type=object_type,
        rules=base_rules,
        quick_share=quick_hare,
    )

    unmatching_shared_objects.append(object_sharing)
    project_settings.get_raw()["exposedObjects"]["objects"] = unmatching_shared_objects

    project_settings.save()


def share_object(
    object_name: str,
    object_type: str,
    current_project: DSSProject,
    target_project_key: str,
    quick_hare=False,
):
    """
    Share an object base on their name and type to a targeted project
    :param: object_name str: Name of the element to share
    :param: object_type str: Type of the element to share (must be in ["DATASET","MANAGED_FOLDER", "SAVED_MODEL"])
    :param: current_project DSSProject: Current project where the element is originated
    :param: target_project_key str: Project key of the target dataset
    :param: quick_hare bool : Is the quick sharing activated for this element
    """
    assert object_type in [
        "DATASET",
        "MANAGED_FOLDER",
        "SAVED_MODEL",
    ], f'{object_type} is not in ["DATASET","MANAGED_FOLDER", "SAVED_MODEL"]'

    if object_type == "MANAGED_FOLDER":
        object_id = get_managed_folder_id(current_project, object_name)
    elif object_type == "SAVED_MODEL":
        object_id = get_deployed_model_id(current_project, object_name)
    else:
        object_id = object_name

    share_object_from_id(
        object_id,
        object_type,
        current_project=current_project,
        target_project_key=target_project_key,
        quick_hare=quick_hare,
    )


def unshare_object_from_id(
    object_id: str,
    object_type: str,
    current_project: DSSProject,
    target_project_key: str,
    quick_hare=False,
):
    """
    Unshare an object base on their name and type from a targeted project
    :param: object_id str: ID of the element to unshare
    :param: object_type str: Type of the element to unshare (must be in ["DATASET","MANAGED_FOLDER", "SAVED_MODEL"])
    :param: current_project DSSProject: Current project where the element is originated
    :param: target_project_key str: Project key of the target dataset
    :param: quick_hare bool : Is the quick sharing activated for this element
    """
    project_settings = current_project.get_settings()
    previous_shared_objects = project_settings.get_raw()["exposedObjects"]["objects"]

    # we check if this specific object is already shared to avoid erasing previous sharing
    matching_shared_objects = list(
        filter(
            lambda dct: (dct["localName"] == object_id)
            and (dct["type"] == object_type),
            previous_shared_objects,
        )
    )

    unmatching_shared_objects = list(
        filter(
            lambda dct: (dct["localName"] != object_id) or (dct["type"] != object_type),
            project_settings.get_raw()["exposedObjects"]["objects"],
        )
    )

    if len(matching_shared_objects) > 0:
        # there's previous sharing
        base_rules = matching_shared_objects[0]["rules"]
        other_project_rules = list(
            filter(lambda dct: dct["targetProject"] != target_project_key, base_rules)
        )
    else:
        # Object is not already shared
        return None

    if len(other_project_rules) > 0:
        # there's still project where this object is shared
        # compute a new dict to replace in the project settings
        object_sharing = compute_object_to_share_settings(
            object_name=object_id,
            object_type=object_type,
            rules=other_project_rules,
            quick_share=quick_hare,
        )
        unmatching_shared_objects.append(object_sharing)

    project_settings.get_raw()["exposedObjects"]["objects"] = unmatching_shared_objects
    project_settings.save()


def unshare_object(
    object_name: str,
    object_type: str,
    current_project: DSSProject,
    target_project_key: str,
    quick_hare=False,
):
    """
    Unshare an object base on their name and type from a targeted project
    :param: object_name str: Name of the element to unshare
    :param: object_type str: Type of the element to unshare (must be in ["DATASET","MANAGED_FOLDER", "SAVED_MODEL"])
    :param: current_project DSSProject: Current project where the element is originated
    :param: target_project_key str: Project key of the target dataset
    :param: quick_hare bool : Is the quick sharing activated for this element
    """

    assert object_type in [
        "DATASET",
        "MANAGED_FOLDER",
        "SAVED_MODEL",
    ], f'{object_type} is not in ["DATASET","MANAGED_FOLDER", "SAVED_MODEL"]'

    if object_type == "MANAGED_FOLDER":
        object_id = get_managed_folder_id(current_project, object_name)
    elif object_type == "SAVED_MODEL":
        object_id = get_deployed_model_id(current_project, object_name)
    else:
        object_id = object_name

    unshare_object_from_id(
        object_id,
        object_type,
        current_project=current_project,
        target_project_key=target_project_key,
        quick_hare=quick_hare,
    )
